// Package file implements filesystem-based storage for MDM services
package file

import (
	"errors"
	"io/ioutil"
	"os"
	"path"

	"github.com/jessepeterson/nanomdm/cryptoutil"
	"github.com/jessepeterson/nanomdm/mdm"
)

const (
	AuthenticateFilename = "Authenticate.plist"
	TokenUpdateFilename  = "TokenUpdate.plist"
	UnlockTokenFilename  = "UnlockToken.dat"
	SerialNumberFilename = "SerialNumber.txt"
	IdentityCertFilename = "Identity.pem"
	DisabledFilename     = "Disabled"

	CertAuthFilename             = "CertAuth.sha256.txt"
	CertAuthAssociationsFilename = "CertAuth.txt"

	// The associations for "sub"-enrollments (that is: user-channel
	// enrollments to device-channel enrollments) are stored in this
	// directory under the device's directory.
	SubEnrollmentPathname = "SubEnrollments"
)

// FileStorage implements filesystem-based storage for MDM services
type FileStorage struct {
	path string
}

// New creates a new FileStorage backend
func New(path string) (*FileStorage, error) {
	err := os.Mkdir(path, 0755)
	if err != nil && !errors.Is(err, os.ErrExist) {
		return nil, err
	}
	return &FileStorage{path: path}, nil
}

type enrollment struct {
	id string
	fs *FileStorage
}

func (s *FileStorage) newEnrollment(id string) *enrollment {
	return &enrollment{fs: s, id: id}
}

func (e *enrollment) dir() string {
	return path.Join(e.fs.path, e.id)
}

func (e *enrollment) mkdir() error {
	return os.MkdirAll(e.dir(), 0755)
}

func (e *enrollment) dirPrefix(name string) string {
	return path.Join(e.dir(), name)
}

func (e *enrollment) writeFile(name string, bytes []byte) error {
	if name == "" {
		return errors.New("write: empty name")
	}
	if err := e.mkdir(); err != nil {
		return err
	}
	return ioutil.WriteFile(e.dirPrefix(name), bytes, 0644)
}

func (e *enrollment) readFile(name string) ([]byte, error) {
	if name == "" {
		return nil, errors.New("write: empty name")
	}
	return ioutil.ReadFile(e.dirPrefix(name))
}

// assocSubEnrollment writes an empty file of the sub (user) enrollment for tracking.
func (e *enrollment) assocSubEnrollment(id string) error {
	subPath := e.dirPrefix(SubEnrollmentPathname)
	if err := os.MkdirAll(subPath, 0755); err != nil {
		return err
	}
	f, err := os.Create(path.Join(subPath, id))
	if err != nil {
		return err
	}
	return f.Close()
}

// listSubEnrollments returns an array of the sub-enrollment IDs
func (e *enrollment) listSubEnrollments() (ids []string) {
	entries, err := os.ReadDir(e.dirPrefix(SubEnrollmentPathname))
	if err != nil && !errors.Is(err, os.ErrNotExist) {
		return nil
	}
	for _, entry := range entries {
		if len(entry.Name()) > 10 {
			ids = append(ids, entry.Name())
		}
	}
	return
}

func (e *enrollment) removeSubEnrollments() error {
	return os.RemoveAll(e.dirPrefix(SubEnrollmentPathname))
}

// StoreAuthenticate stores the Authenticate message
func (s *FileStorage) StoreAuthenticate(r *mdm.Request, msg *mdm.Authenticate) error {
	e := s.newEnrollment(r.ID)
	if r.Certificate != nil {
		if err := e.writeFile(IdentityCertFilename, cryptoutil.PEMCertificate(r.Certificate.Raw)); err != nil {
			return err
		}
	}
	// A nice-to-have even though it's duplicated in msg
	if msg.SerialNumber != "" {
		err := e.writeFile(SerialNumberFilename, []byte(msg.SerialNumber))
		if err != nil {
			return err
		}
	}
	return e.writeFile(AuthenticateFilename, []byte(msg.Raw))
}

// StoreTokenUpdate stores the TokenUpdate message
func (s *FileStorage) StoreTokenUpdate(r *mdm.Request, msg *mdm.TokenUpdate) error {
	e := s.newEnrollment(r.ID)
	// the UnlockToken should be saved separately in case future
	// TokenUpdates do not contain it and it gets overwritten
	if len(msg.UnlockToken) > 0 {
		if err := e.writeFile(UnlockTokenFilename, msg.UnlockToken); err != nil {
			return err
		}
	}
	if r.ParentID != "" {
		parentEnrollment := s.newEnrollment(r.ParentID)
		if err := parentEnrollment.assocSubEnrollment(r.ID); err != nil {
			return err
		}
	}
	if err := e.writeFile(TokenUpdateFilename, []byte(msg.Raw)); err != nil {
		return err
	}
	// delete the disabled flag to let signify this enrollment is enabled
	if err := os.Remove(e.dirPrefix(DisabledFilename)); err != nil && !errors.Is(err, os.ErrExist) {
		return err
	}
	return nil
}

func (s *FileStorage) Disable(r *mdm.Request) error {
	if r.ParentID != "" {
		return errors.New("can only disable a device channel")
	}
	// assemble list of IDs for which to disable
	e := s.newEnrollment(r.ID)
	disableIDs := e.listSubEnrollments()
	disableIDs = append(disableIDs, r.ID)
	// disable all of the ids
	for _, id := range disableIDs {
		e := s.newEnrollment(id)
		// write zero-byte disabled marker
		if err := e.writeFile(DisabledFilename, nil); err != nil {
			return err
		}
	}
	return e.removeSubEnrollments()
}
